/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.mediator;

import org.xquark.xml.xdbc.XMLConnection;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xml.xdbc.XMLDriverManager;

public class SubAccessor {
	// **********************************************************************
	// * VERSIONING
	// **********************************************************************
	private static final String RCSRevision = "$Revision: 1.7 $";
	private static final String RCSName = "Name";

	// **********************************************************************
	// * CLASS VARIABLE
	// **********************************************************************
	private String str_name;
	private String str_driver;
	private String str_connection;
	private boolean isConnected;
	private XMLConnection connection = null;
	private Mediator mediator = null;

	private String login = null;
	private String passwd = null;

	// ************************************************************************
	// * INITIALISATION
	// ************************************************************************
	/**
	 *
	 */
	public SubAccessor(Mediator mediator, String name, String driver, String connection) throws MediatorException {
		this.str_name = name;
		this.str_driver = driver;
		this.str_connection = connection;

		this.mediator = mediator;

		this.isConnected = false;
	}

	/**
	 *
	 */
	public SubAccessor(Mediator mediator, String name, String driver, String connection, String login, String passwd) throws MediatorException {
		this.str_name = name;
		this.str_driver = driver;
		this.str_connection = connection;
		this.login = login;
		this.passwd = passwd;

		this.mediator = mediator;

		this.isConnected = false;
	}

	// ************************************************************************
	// * CONNECTION/DISCONNECTION METHODS
	// ************************************************************************
	/**
	 * Load the accessor driver and open a connection.
	 */
	public XMLConnection getConnection() throws MediatorException {
		return getConnection(false);
	}
	public XMLConnection getConnection(boolean forceNew) throws MediatorException {
		XMLConnection foundConnection = null;
		try {
			if (connection == null || forceNew) {
				Class.forName(str_driver);
				if ((login != null) && (passwd != null))
					foundConnection = XMLDriverManager.getConnection(str_connection, login, passwd);
				else
					foundConnection = XMLDriverManager.getConnection(str_connection);
				if (foundConnection != null)
					isConnected = true;
				if (connection == null)
					connection = foundConnection;
			} else
				foundConnection = connection;
			return foundConnection;
		} catch (XMLDBCException e) {
			throw new MediatorException("Could not get a connection for " + str_name + ": " + e.getMessage(), e);
		} catch (ClassNotFoundException e) {
			throw new MediatorException("Could not get a connection for " + str_name + ": " + e.getMessage(), e);
		}
	}

	/**
	 * Close the accessor's connection.
	 */
	public void close() throws MediatorException {
		try {
			if (connection != null)
				connection.close();
			isConnected = false;
		} catch (XMLDBCException xmldbce) {
			throw new MediatorException("SubAccessor.close: " + xmldbce.getMessage());
		}
	}

	// ************************************************************************
	// * GET/SET METHODS
	// ************************************************************************
	/**
	 * Return the name of this accessor.
	 */
	public String getName() throws MediatorException {
		return str_name;
	}
	public String getStrDriver() throws MediatorException {
		return str_driver;
	}
	public String getStrConnection() throws MediatorException {
		return str_connection;
	}
	public String getLogin() throws MediatorException {
		return login;
	}
	public String getPassword() throws MediatorException {
		return passwd;
	}
	public boolean getState() throws MediatorException {
		return isConnected;
	}
}
